#ifndef IN_TS_H
#define IN_TS_H

		   /******************************************
		   *		     TS.H		     *
	    	   * Copyright TimeSlice, Inc. 1985, 86, 87. *
		   ******************************************/

#define MICROSOFT
#define ML

#ifdef TURBOC
#define TCVARGS	...
#endif
#ifndef TURBOC
#define TCVARGS
#endif



/***
* PROCESS
* Each task has one of these structure (returned by spawn()), to hold
* task specific information.
***/
typedef struct process {	/*process information*/
  struct process *nxt,		/*pointer to next process*/
		 *crnxt,	/*next critical process waiting*/
		 *chnxt;	/*next process waiting for a channel*/	
  char	*name,			/*process name */
	*usrptr;		/*reserved for your application*/
  int	crwait,			/*critical class being waited for*/
	crlev,			/*nesting level of crit blocks*/
	chlev,			/*nesting level of chan request*/
  	tunits,			/*number of consecutive time units*/
	status,			/*status flags*/
	event,			/*value set by an asynchronous interrupt*/
	psp,			/*process' stack pointer*/
	pss,			/*process' stack segment*/
				/***This must always be last***/
	regsfpc;		/*beginning of FPC save/restore area*/
} PROCESS;

extern PROCESS  *curproc;	/*pointer to current process*/
extern PROCESS	*_mainproc;	/*pointer to TSMAIN	    */
extern PROCESS	*_idleproc;	/*pointer to TSMAIN	    */
extern PROCESS	*spawn( char *, int, void (*)(), ),
				/*process spawning function*/
	      	*getproc() ;	/*get PROCESS pointer from name*/	

/***
* CRCLASS
* Each critical class is defined by such an entry in crit_tab[]
***/
typedef struct {
  PROCESS *crcur,
	  *crblkd,
	  *crrear;
  int	  crlev;
} CRCLASS;

extern CRCLASS	crit_tab[],
		*crit_stop;

#define DOS_CRCLASS 0			/*critical for any DOS service	*/
#define EMS_CRCLASS 1			/*critical EMS buffer control	*/
#define EMS_emscall 2			/*critical for any EMS call	*/

/***
* WPROC
* To be used for event counting.
***/
typedef struct {			/*information used by icntr()*/
  unsigned long	reset,			/* reset value of counter */
		cntr;			/* current value of counter */
  PROCESS *pptr ;			/* process to be woken up when !cntr */
} WPROC ;

/***
* TIMER
* To be used by TIMER function FTIMER(), PTIMER() and RMVTIMER() .
***/
typedef struct timer {
  struct timer *nxt;
  unsigned long	reset,			/* reset value of counter */
		cntr;			/* current value of counter */
  void (*func)();			/* function to call when !cntr */
  char *arg;				/* argument to pass to func*/
} TIMER;

/*****
* SHIFT
* This structure contains the shift status value that will trigger
* the function pointer func.
*****/
typedef struct shift {
  struct shift *nxt;
  void	(*func)() ;			/* function pointer to be executed..*/
  char	status,				/* ..when kbd status equals this status */
	infunc;				/* 1 -> already in use*/
} SHIFT ;

/***
* Various global variables
***/
extern unsigned char
	  *_DosFlag;			/*slicer interrupted DOS ? */

extern unsigned int
	   _stkseg,			/*main stack segment*/
	   _codseg,			/*main code segment*/
	   _datseg,			/*static data segment*/
	   _vidseg,			/*video segment used by prout*/
	   _i13Flag;			/*slicer interrupted int13 */

extern int _timcnt,			/*time units left for curproc*/
	   _sslev,			/*cur. level of switcher sleepiness*/
	   _tckcnt,			/*tick counter used by slicer*/
	   _tckrst;			/*tick counter reset*/


#define TS_TASK   void			/* task type */
#define FLAGS	0xF202			/*starting flags of process*/

		/***STATUS VALUES***/
#define READY		0x0000		/*process ready to run*/
#define	ASLEEP		0x0001		/*regular suspended*/
#define CRITWAIT	0x0002		/*waiting to be critical*/
#define CHANWAIT	0x0004		/*waiting for channel to be available*/
#define KILL		0x0008		/*If this bit is set then the process is killed by critend()*/
#define SAVEFPC		0x0010		/*if float. coproc. should be saved & restored*/
#define DOSCRIT 	0x0100		/*Main Task, Overlay loader etc 	   */
#define INDOS		0x0200		/*Main Task in DOS		    */

/***
* Shift macros for ONKEY()
***/
#define	R_SHIFT		1
#define	L_SHIFT		2
#define	CTRL		4
#define	ALT		8
#define SCRLLCK 	0x10
#define	NUMLCK		0x20
#define	CAPSLOCK	0x40

/***
* Macros required for lattice compatibility
***/
#if defined(MICROSOFT) || defined(TURBOC)
#define getmem malloc
#define	rlsmem free

#define	movmem(from,to,n) memcpy(to,from,n)
#define setmem(to,cnt,c) memset(to,c,cnt)

#define O_RAW	O_BINARY


#define max(a,b)	(((a) > (b)) ? (a) : (b))
#define min(a,b)	(((a) < (b)) ? (a) : (b))
#endif


#define CW_TOP	  0
#define CW_AHEAD  1
#define CW_ACK	  2
#define CW_NAK	  3
#define CW_RESET  3
#define CW_FLUSH  4
#define CW_APPEND 5
#define CW_BOT	  6


/***
* QUEUE STRUCTURE
* The Queue is a byte oriented FIFO buffer. It uses two indexes to know
* where to put bytes in the queue (top), and where to get bytes from the
* queue (bot). The functions in charge of putting/getting bytes on/from
* the queue must do 3 things:
* 	1> whenever an index reaches size it is reset to zero,
*	2> if top/bot reaches bot/top set the full/empty flag,
*	3> when putting/getting a byte reset the empty/full flags.
***/
typedef struct {
  short top,			/*index of next spot available to put a byte*/
	ahead,			/*index of next byte for write ahead  */
	bot,			/*index of next byte available*/
	size;			/*physical size (in bytes) of buffer*/
  char  *buf,			/* pointer to physical buffer */
  	full,			/*set when top reaches bot*/
	empty;			/*set when bot reaches top*/
} QUEUE;


/***
* CHAN
* Structure used for interprocess communication 
***/
#define	CNSIZ	8			/* size of channel name */

typedef struct chan {
    struct chan	*nxt ;			/* Pointer to next structure */
    char cname[ CNSIZ + 1 ] ;		/* name assigned to channel */
    int	openlev ;			/* open level count */
    PROCESS *rf ,			/* NULL - okay to read */
	    *wf ;			/* NULL - okay to write */
    QUEUE q ;				/* queue structure */
} CHAN ;

extern CHAN *_chfrnt;


/***
* TS_MAIN
* This macro insures that the function main is actually called ts_main.
* We provide the main ourselves where the slicer is installed.
***/
#define main	ts_main
#define exit	exitmp
#define cli	_cli


/***
* FARPTR
* typedef used to represent SEG:OFF ptrs
***/
typedef union {
  char *ptr;
  long l;
  struct {
    int off,
    	seg;
  }i;
} FARPTR;

/***
* INTLST - This structure is the entry point of the interrupt CODE, and
* also its exit point. The code in this structure calls the interrupt manager
* which in turn calls the attached function.
***/
typedef struct intlst {
  struct intlst *nxt,		/*ptr to next intlst node*/
		*inxt;			/*ptr to next intlist node with same intno*/  
  int	stksiz;			/*stack size required by func*/
  char  *arg;			/*argument passed by user */
  long  iid;			/*interrupt func'c id*/
  short	intno;			/*interupt processed*/
  int	(*func)();		/*function attached to the interrupt*/
  char	dummy1,			/*must fill hole so that imoff is even*/
        opcall;			/*'call far' opcode 9EH*/
  int   imoff,
  	imseg;			/*ptr to interupt manager*/
  char	dummy2,			/*must fill hole (filled with NOP)*/
	opjmp;			/*'jmp far' opcode EAH (or IRET)*/
  FARPTR jmpaddr;		/*original value of interrupt vector*/
} INTLST;

/***
* INTSTATE - This structure maps the stack of the interrupted code
* where the cpu registers and flags are pushed.
***/
typedef struct {
  int	es, ds, 		/* iAPX 86/88 registers */
  	di, si,
	sp, bp;
  union {
    struct {
      int bx, dx,
       	  cx, ax;
    } w;
    struct {
      char bl, bh, dl, dh,
           cl, ch, al, ah;
    } b;
  } r;
  int	imrip, 			/* return ip value of intmgr */
	imrcs, 			/* return cs value of intmgr */
	intip, 			/* return ip value of interrupted code */
	intcs, 			/* return cs value of interrupted code */
	flags ;			/* cpu flags of interrupted code */
} INTSTATE ;

#define	BEFORE	1		/*execute ISR to attach BEFORE original ISR*/
#define REPLACE	0		/*replace original ISR with the one to attach*/

extern INTLST *ilhead;		/*head of INTLST linked list*/
extern long	intatt( int, int, void (*)(), int, );
				/* interrupt attach function */
#define CALLFAR	0x9A		/*opcode for 'call far'*/
#define JUMPFAR 0xEA		/*opcode for 'jump far'*/
#define IRET	0xCF		/*opcode for 'iret'*/
#define	NOP	0x90		/*opcode for NOP*/

extern void (*intexit)();	/*if set, points to intrest()*/

/***
* ISRJMPON(IP) - This macro is called by an ISR in order to jump to the 
* previous ISR for that I.V. IP is a pointer to the ISR INTLST.
***/
#define	isrjmpon(ip)	((ip)->opjmp=JUMPFAR)

/***
* ISRJMPOFF(IP) - This macro is called by an ISR in order to not jump to the 
* previous ISR for that I.V. IP is a pointer to the ISR INTLST.
***/
#define	isrjmpoff(ip)	((ip)->opjmp=IRET)

/***
* SSLEEP() and SWAKE()
***/
#define	ssleep()	(_sslev++)	/*increase level of slicer sleep*/
#define swake()		(_sslev--)	/*decrease level of slicer sleep*/

/***
* Definitions and declarations specific to the microprocessor or machine used.
* This version is for the IBM and true compatibles.
***/
#define VIDADDR	0xB000		/*monochrome video segment*/
#define COLORVID 0xB800		/*color video segment*/
#define	REVVID	0x70		/*reverse video attribute*/
#define NORMVID	0x07		/*normal video attribute*/
#define BLDVID	0x18		/*bold video attribute*/
#define ULNVID	0x01		/*underline video attribute*/ 

#define HARDKBD	  0x09		/*hardware keyboard interrupt*/

/***
* CRITICAL FUNCTIONALITY
* These macros allow a function which does not return an integer to be used
* with an equivalent of crit().
***/
typedef	long (*PFL)( int, long (*)(), TCVARGS);
typedef	double (*PFD)( int, double (*)(), TCVARGS);
typedef char  *(*PFCP)( int,TCVARGS);
typedef char  (*PFC)( int, char (*)(), TCVARGS);

extern 	PFL  _lcrit ;
extern  PFD  _dcrit ;
extern	PFCP _cpcrit ;
extern	PFC _ccrit ;

#define	lcrit	(*_lcrit)	/* To be used for executing critical functions returning longs */
#define	dcrit	(*_dcrit)	/* To be used for executing critical functions returning doubles */
#define	cpcrit	(*_cpcrit)	/* To be used for executing critical functions returning character pointers */
#define ccrit   (*_ccrit)	/* To be used for executing critical functions returning character */

extern PROCESS *_crfrnt, *_crrear;

/***
* floating point coprocessor processing
***/
extern char _tsfpc;		/*0 -> no fpc, 1 -> fpc present*/
extern char _fpc_support;	/*0 -> ignore fpc*/

/*TimeSlicer Function Prototypes*/
int	 cli(void);
void	 critend( int );
void	 critstart( int );
void	 endself();
void	 exitmp( int );
int	 getf();
PROCESS	*getproc( char * );
void	 guc(void);
char	 kbdget(void);
int	 kbdscan(void);
void	 prout( int, int, int, char * );
void	 putf( int );
void	 wakeproc( PROCESS *, int );
int	 cclose( CHAN *);
int	 cclosbuf( CHAN *);
CHAN	*copen( char *, int);
CHAN	*copenbuf( char *, CHAN *, int);
int	 _cread( CHAN *, int, char*, int );
void	 _creadw( CHAN *, int, char*, int );
int	 _cwrite( CHAN *, int, char*, int );
void	 _cwritew( CHAN *, int, char*, int );
short	 cwset( CHAN *, short);
void	 crel( CHAN *);
short	 cnfill( CHAN *, short );
short	 cncount( CHAN *, short );
void	 creq( CHAN *, int );
void	 icntr( INTLST *);
INTLST	*intdet( long );
void	 intrest();
char	*_ts_malloc( int );
void	 ts_rrip();
int	 kill(PROCESS *);
void	 tickrate( int );
long	 timestamp( long * );
long	 elapsed( long * );
SHIFT	*onshift( int, void (*)() );
void	 rmvshift( SHIFT * );
TIMER	*ptimer( unsigned long, PROCESS * );
void	 rmvtimer( TIMER * );
void	 reset_timer(TIMER *tp, unsigned long tunits);


#ifndef TURBOC
PROCESS	*spawn( char *, int, void (*)(), );
int	 crit( int, int (*)(), );
long	 intatt( int, int, void (*)(), int, );
TIMER	*ftimer( unsigned long, void (*)(), );
#endif
#ifdef TURBOC
int	 crit( int, int (*)(),TCVARGS);
PROCESS	*spawn();
long	 intatt();
TIMER	*ftimer();
#endif

#endif
